from enum import Enum
from .note import Note

KEYBOARD = "CxDxEFxGxAxB"
WHITE_KEYS = "CDEFGAB"


class Interval:

    def __init__(self, number: int, quality):
        self.__number__ = number
        self.__quality__ = quality

    def number(self):
        return self.__number__

    def quality(self):
        return self.__quality__

    def semitones(self):
        return self.INTERVAL_SEMITONES[self.__number__][self.__quality__]

    def counted_from(self, note: Note) -> Note:
        transposedName = WHITE_KEYS[(number_in_c(note) + self.number() - 1) % 7]
        diatonicDistance = (KEYBOARD.index(transposedName) - KEYBOARD.index(note.name())) % 12
        transposedModifier = note.accidental_modifier() + self.semitones() - diatonicDistance
        return Note.from_name_and_modifier(transposedName, transposedModifier)

    @classmethod
    def between(cls, fromNote: Note, toNote: Note):
        number = number_between(fromNote, toNote)
        semitones = semitones_between(fromNote, toNote)
        return Interval.from_number_and_semitones(number, semitones)

    @classmethod
    def from_number_and_semitones(cls, number: int, semitones: int) -> 'Interval':
        return Interval(number, cls.INTERVAL_QUALITIES[number][semitones])

    class Quality(Enum):
        PERFECT = "P"
        AUGMENTED = "A"
        DIMINISHED = "d"
        MAJOR = "M"
        MINOR = "m"

    @classmethod
    def unison(cls, quality: Quality = Quality.PERFECT):
        return Interval(1, quality)

    @classmethod
    def second(cls, quality: Quality = Quality.MAJOR):
        return Interval(2, quality)

    @classmethod
    def third(cls, quality: Quality = Quality.MAJOR):
        return Interval(3, quality)

    @classmethod
    def fourth(cls, quality: Quality = Quality.PERFECT):
        return Interval(4, quality)

    @classmethod
    def fifth(cls, quality: Quality = Quality.PERFECT):
        return Interval(5, quality)

    @classmethod
    def sixth(cls, quality: Quality = Quality.MAJOR):
        return Interval(6, quality)

    @classmethod
    def seventh(cls, quality: Quality = Quality.MINOR):
        return Interval(7, quality)

    INTERVAL_SEMITONES = {
        1: {
            Quality.PERFECT: 0,
            Quality.AUGMENTED: 1
        },
        2: {
            Quality.DIMINISHED: 0,
            Quality.MINOR: 1,
            Quality.MAJOR: 2,
            Quality.AUGMENTED: 3
        },
        3: {
            Quality.DIMINISHED: 2,
            Quality.MINOR: 3,
            Quality.MAJOR: 4,
            Quality.AUGMENTED: 5
        },
        4: {
            Quality.DIMINISHED: 4,
            Quality.PERFECT: 5,
            Quality.AUGMENTED: 6
        },
        5: {
            Quality.DIMINISHED: 6,
            Quality.PERFECT: 7,
            Quality.AUGMENTED: 8
        },
        6: {
            Quality.DIMINISHED: 7,
            Quality.MINOR: 8,
            Quality.MAJOR: 9,
            Quality.AUGMENTED: 10
        },
        7: {
            Quality.DIMINISHED: 9,
            Quality.MINOR: 10,
            Quality.MAJOR: 11,
            Quality.AUGMENTED: 12
        }
    }

    INTERVAL_QUALITIES = {
        1: {
            0: Quality.PERFECT,
            1: Quality.AUGMENTED
        },
        2: {
            0: Quality.DIMINISHED,
            1: Quality.MINOR,
            2: Quality.MAJOR,
            3: Quality.AUGMENTED
        },
        3: {
            2: Quality.DIMINISHED,
            3: Quality.MINOR,
            4: Quality.MAJOR,
            5: Quality.AUGMENTED
        },
        4: {
            4: Quality.DIMINISHED,
            5: Quality.PERFECT,
            6: Quality.AUGMENTED
        },
        5: {
            6: Quality.DIMINISHED,
            7: Quality.PERFECT,
            8: Quality.AUGMENTED
        },
        6: {
            7: Quality.DIMINISHED,
            8: Quality.MINOR,
            9: Quality.MAJOR,
            10: Quality.AUGMENTED
        },
        7: {
            9: Quality.DIMINISHED,
            10: Quality.MINOR,
            11: Quality.MAJOR,
            12: Quality.AUGMENTED,
            0: Quality.AUGMENTED
        }
    }


def semitones_between(fromNote: Note, toNote: Note) -> int:
    fromPosition = KEYBOARD.index(fromNote.name()) + fromNote.accidental_modifier()
    toPosition = KEYBOARD.index(toNote.name()) + toNote.accidental_modifier()
    return (toPosition - fromPosition) % 12


def number_between(fromNote: Note, toNote: Note) -> int:
    return (number_in_c(toNote) - number_in_c(fromNote)) % 7 + 1


def number_in_c(note: Note) -> int:
    return WHITE_KEYS.index(note.name())



from songpy.music import Interval, Note
from songpy.music.chord import Chord
from songpy.song import Song

PREFERRED_ENHARMONY = {
    Chord.Mode.MINOR: [
        "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "Bb", "B"
    ],
    Chord.Mode.MAJOR: [
        "C", "Db", "D", "Eb", "E", "F", "F#", "G", "Ab", "A", "Bb", "B"
    ]
}


def transpose_by_semitones(song: Song, semitones: int):
    semitones = semitones % 12
    if semitones == 0:
        return song
    first_chord = find_first_chord(song)
    if first_chord is None:
        return song
    interval = preferred_interval(first_chord, semitones)
    song.transpose(interval)
    return song



def find_first_chord(song: Song):
    for part in song.content:
        for line in part.content:
            for atom in line.content:
                if atom.chord is not None:
                    return atom.chord
    return None


def distance_from_c(note: Note) -> int:
    return Interval.between(Note.from_symbol("C"), note).semitones()


def preferred_interval(initial_chord: Chord, semitones: int) -> Interval:
    semitones_from_c = (distance_from_c(initial_chord.root()) + semitones) % 12
    return Interval.between(initial_chord.root(),
                            Note.from_symbol(PREFERRED_ENHARMONY[initial_chord.mode()][semitones_from_c]))



import json
import re

from songpy.music.chord import Chord
from songpy.song import OneChordAtom, SongPart, Song, SongLine, SongBook
from songpy.utils import PolishNoteNamesPreprocessor
from songpy.utils.transpose import transpose_by_semitones

PREPROCESSORS_CHAIN = []
SONG_PART_MARKERS = [
    "^[0-9]\\.\\s*",
    "^ref(ren|((\\.)?)):?\\s*"
]


def register_preprocessor(preprocessor):
    PREPROCESSORS_CHAIN.append(preprocessor)


def is_marked(dictionary: dict) -> bool:
    for marker_pattern in SONG_PART_MARKERS:
        if bool(re.search(marker_pattern, dictionary['content'][0], flags=re.IGNORECASE)):
            return True
    return False


def strip_markers(line: str) -> str:
    result = line
    for marker_pattern in SONG_PART_MARKERS:
        result = re.sub(marker_pattern, "", result, flags=re.IGNORECASE)
    return result


def preprocess_line(raw: str) -> str:
    product = raw
    for preprocessor in PREPROCESSORS_CHAIN:
        product = preprocessor.process(product)
    return product


def song_part_from_dictionary(dictionary: dict) -> SongPart:
    raw_lines = dictionary['content'].strip().split('\n')
    should_be_marked = is_marked({'content': raw_lines})
    if should_be_marked:
        raw_lines[0] = strip_markers(raw_lines[0])
    lines = list(map(parse_line, raw_lines))
    return SongPart(content=lines, type=dictionary['type'], marked=should_be_marked)


def parse_line(raw: str) -> [OneChordAtom]:
    product = preprocess_line(raw)
    pattern = r'(\[([^\]]*)\])?' + r'([^\[]*)'
    matches = re.finditer(pattern, product)
    return SongLine([OneChordAtom(text=match.group(3), chord=Chord.from_symbol(match.group(2)))
                     for match in matches][:-1])


def song_from_dictionary(dictionary: dict) -> Song:
    output = Song(dictionary['title'])
    output.lyrics_author = dictionary['lyrics']
    output.music_author = dictionary['music']
    # output.translation_author = dictionary['translated']
    content = dictionary['content']
    for song_part in content:
        output.add_song_part(song_part_from_dictionary(song_part))
    if 'transposition' in dictionary:
        output = transpose_by_semitones(output, dictionary['transposition'])
    if 'truncation' in dictionary:
        output.content = output.content[:dictionary['truncation']]
    return output


def songbook_from_dictionary(dictionary: dict) -> SongBook:
    output = SongBook()
    song_dicts = dictionary
    for song_dict in song_dicts:
        output.songs.append(song_from_dictionary(song_dict))
    output.sort_songs()
    return output


class SimplePrinter:
    @classmethod
    def print(cls, song: Song) -> str:
        output = song.title
        verse_count = 1
        for part in song.content:
            output += "\n"
            prefix = ""
            if part.marked:
                if part.type == 'chorus':
                    prefix = "Ref.: "
                if part.type == 'verse':
                    prefix = str(verse_count) + ". "
                    verse_count += 1
                output += prefix
            for line in part.content:
                output += line.lyrics() + " | " + " ".join(list(map(simple_printer, line.chords())))
                output += "\n"
        return output


def simple_printer(obj) -> str:
    return str(obj)


if __name__ == '__main__':
    register_preprocessor(PolishNoteNamesPreprocessor())
    file = open("song.json", 'r')
    song = json.load(file)
    file.close()
    song = song_from_dictionary(song)
    print(SimplePrinter().print(song))



import re


class PolishNoteNamesPreprocessor:
    @classmethod
    def process(cls, raw: str) -> str:
        product = re.sub(r"\[B", "[Bb", raw)
        product = re.sub(r"\[H", "[B", product)
        return product



class OneChordAtom:
    def __init__(self, text: str, chord):
        self.chord = chord
        self.text = text

    def transpose(self, interval):
        if self.chord is not None:
            self.chord = self.chord.transpose(interval)


class SongLine:
    def __init__(self, content: [OneChordAtom]):
        self.content = content

    def lyrics(self) -> str:
        return "".join([atom.text for atom in self.content])

    def chords(self):
        return [atom.chord for atom in self.content if atom.chord is not None]

    def formatted_chords(self, chord_printer):
        return " ".join([chord_printer(atom.chord) for atom in self.content if atom.chord is not None])

    def transpose(self, interval):
        for atom in self.content:
            atom.transpose(interval)


class SongPart:
    def __init__(self, content: [], type: str = 'verse', marked: bool = False):
        self.content = content
        self.type = type
        self.marked = marked

    def lyrics(self):
        return '\n'.join(self.lines())

    def lines(self):
        return [
            line.lyrics()
            for line in self.content
        ]

    def transpose(self, interval):
        for line in self.content:
            line.transpose(interval)


class Song:
    def __init__(self, title):
        self.title = title
        self.lyrics_author = None
        self.music_author = None
        self.translation_author = None
        self.content = []

    def add_song_part(self, songpart):
        self.content.append(songpart)

    def transpose(self, interval):
        for part in self.content:
            part.transpose(interval)


class SongBook:
    def __init__(self):
        self.songs = []

    def sort_songs(self):
        self.songs.sort(key=SongBook.compare_function)

    @staticmethod
    def compare_function(e):
        return e.title

    @staticmethod
    def from_dict(dictionary):
        output = SongBook()
        song_dicts = dictionary
        for dict in song_dicts:
            output.songs.append(Song.from_dict(dict))
        output.sort_songs()
        return output



import re
from enum import Enum

from songpy.music import Note, Interval


class Chord:

    def __init__(self, root, mode):
        self.__mode__ = mode
        self.__root__ = root
        self.__symbol__ = self.__str__()

    def mode(self):
        return self.__mode__

    def root(self) -> Note:
        return self.__root__

    def transpose(self, interval: Interval):
        return Chord(interval.counted_from(self.__root__), self.mode())

    @classmethod
    def from_symbol(cls, symbol: str):
        if symbol is None:
            return None
        pattern = r'([a-gA-G](#+|b+)?)([mM]?).*'
        match = re.search(pattern, symbol)
        root = Note.from_symbol(match.group(1))
        mode = cls.Mode.of(match.group(3))
        return Chord(root, mode)

    def __eq__(self, other):
        if not isinstance(other, Chord):
            return False
        return self.__symbol__ == other.__symbol__

    class Mode(Enum):
        MINOR = "m"
        MAJOR = ""

        @classmethod
        def of(cls, string):
            return Chord.Mode(string.lower())

    def __str__(self):
        return str(self.__root__) + self.__mode__.value


if __name__ == '__main__':
    pass



import os

from setuptools import setup, find_packages


version = os.getenv("PACKAGE_VERSION", "0.0.3a1")


setup(
    name='songpy',
    version=version,
    description='Package with objects representing objects of musical notation',
#    long_description=readme,
    author='Karol Ważny',
#    author_email='me@kennethreitz.com',
#    url='https://github.com/kennethreitz/samplemod',
#    license=license,
    packages=find_packages(exclude=('tests', 'docs'))
)



class Note:
    @classmethod
    def from_symbol(cls, symbol: str):
        return Note(symbol)

    @classmethod
    def from_name_and_modifier(cls, letterName: str, modifier: int):
        accidental = cls.__accidental_from_modifier__(modifier)
        return Note(letterName + accidental)

    def __init__(self, symbol: str):
        self.__letter_name__ = symbol[0].upper()
        self.__accidental_modifier__ = self.__accidental_modifier_from_accidental__(symbol[1:].lower())

    def name(self) -> str:
        return self.__letter_name__

    def accidental_modifier(self) -> int:
        return self.__accidental_modifier__

    def accidental(self) -> str:
        return self.__accidental_from_modifier__(self.__accidental_modifier__)

    def symbol(self) -> str:
        return self.name() + self.accidental()

    @staticmethod
    def __accidental_from_modifier__(modifier: int) -> str:
        symbol = "#"
        if modifier < 0:
            symbol = "b"
        return abs(modifier) * symbol

    @staticmethod
    def __accidental_modifier_from_accidental__(accidental: str) -> int:
        direction = 1
        if accidental.__contains__("b"):
            direction = -1
        return direction * len(accidental)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (self.__letter_name__ == other.__letter_name__ and
                self.__accidental_modifier__ == other.__accidental_modifier__)

    def __str__(self):
        return self.symbol()



import songpy.music.note


class PolishNotesNamesFormatter:
    sharp = "is"
    flat = "es"
    dropped_flat_suffix_letters = {
        "C": 0,
        "D": 0,
        "E": 1,
        "F": 0,
        "G": 0,
        "A": 1,
        "B": 2
    }

    def format(self, note: songpy.music.note.Note):
        return self.base_name(note) + self.accidental_suffix(note)

    def base_name(self, note: songpy.music.note.Note):
        note_name = note.name()
        if note_name == "B" and note.accidental_modifier() >= 0:
            note_name = "H"
        return note_name

    def accidental_suffix(self, note: songpy.music.note.Note):
        modifier = note.accidental_modifier()
        if modifier > 0:
            return self.sharp_suffix(modifier)
        return self.flat_suffix(note)

    def sharp_suffix(self, semi_tones: int):
        return semi_tones * self.sharp

    def flat_suffix(self, note: songpy.music.note.Note):
        regular_suffix = - note.accidental_modifier() * self.flat
        first_index = self.dropped_flat_suffix_letters[note.name()]
        return regular_suffix[first_index:]
